/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *     device.c  Created APR 7, 1999     Konstantyn V. Prokopenko
 *                                       (kporoko@analogic.com)
 *    Modified APR 15, 1999               Richard B. Johnson
 *                                        (rjohnson@analogic.com)
 * 
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif
#define __DRIVER__
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "device.h"           /* Our constants, protypes, etc. */

#define NoDEBUG
#ifdef  DEBUG
#define DEB(f) f 
#else
#define DEB(f)
#endif

#define YIELD  \
{                                   \
    current->policy |= SCHED_YIELD; \
    schedule();                     \
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Function prototypes.
 */
static loff_t device_seek(struct file *, loff_t, int);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);
static int device_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static int device_open(struct inode *, struct file *); 
static int device_close(struct inode *, struct file *); 
static void device_isr(int, void *, struct pt_regs *);
int __init init_module(void);
void cleanup_module(void);

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

static const char devname[]="Analogic(tm) Platinum Driver";

static spinlock_t device_lock = SPIN_LOCK_UNLOCKED;

MODULE_AUTHOR("Richard B, Johnson");
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* 
 *   File operation structure contains basic functions to operate a device
 */
static struct file_operations device_fops = {
#if defined(THIS_MODULE)
    THIS_MODULE,
#endif 
    device_seek,                      /* llseek      */
    device_read,                      /* read        */
    device_write,                     /* write       */
    NULL,                             /* readdir     */
    NULL,                             /* poll        */
    device_ioctl,                     /* ioctl       */
    NULL,                             /* mmap        */
    device_open,                      /* open        */
    NULL,                             /* flush       */
    device_close,                     /* close       */
    0, };

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is device-specific information of the kind you specify. The
 *  kernel doesn't care about its contents, only that it exists.
 */

typedef struct {
    const char *dev;          /* Device name           */
    int major;                /* Device major number   */
    int irq;                  /* Interrupt to use      */
    size_t intrs;             /* Number of interrupts  */
    size_t cntrl;             /* Device control bits   */
    size_t open;              /* Device open count     */
    char *buf;                /* Allocated data buffer */
    } INFO;

static INFO *info;

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is how we read from the device.
 */
static ssize_t device_read(struct file *fp, char *buf, size_t count, loff_t *ppos)
{
    DEB(printk("%s read\n", info->dev)); 
    if(buf == NULL) return -EINVAL;
    if(!count) return count;
    if(!access_ok(VERIFY_WRITE, buf, count))     /* Plug a security hole     */
        return -EFAULT;
    info->cntrl |= READ_ENB;                     /* Enable port reads        */
    outb(info->cntrl, LPT_CTRL_PORT);            /* To the control port      */
    YIELD; 
    if(put_user((char)inb(LPT_DATA_PORT), buf)) return -EFAULT;
    return 1;                                    /* Always only 1 byte       */
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is how we write to the device.
 */
static ssize_t device_write(struct file *fp, const char *buf, size_t count, loff_t *ppos)
{
    ssize_t retval;
    char c;

    DEB(printk("%s write\n", info->dev));
    if(buf == NULL) return -EINVAL;
    if(!count) return (ssize_t) count;
    if(!access_ok(VERIFY_READ, buf, count))     /* Plug a security hole     */
        return -EFAULT;
    info->cntrl &= ~READ_ENB;                   /* Set to write only mode     */
    outb(info->cntrl, LPT_CTRL_PORT);           /* To the control port        */
    YIELD;
    retval = count;                             /* We should return count val */
    while(!!(count--))                            
    {
        if(get_user(c, buf++)) return -EFAULT;
        outb(c, LPT_DATA_PORT);                 /* Send all bytes to the port */
    }
    return retval;
}

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Note spinlocks are put around variables that could change while they
 *  are being read. The values are copied into a local variable under the
 *  lock. The lock is released, then the variable is copied to the user.
 *  This is essential. You can't copy to the user under a lock because the
 *  user's data might not be present at this instant so the kernel page-
 *  fault handler will have to get it. You can't page-fault with interrupts
 *  off.
 */
static int device_ioctl(struct inode *inp, struct file *fp, unsigned int cmd, unsigned long arg)
{
    unsigned long flags;
    int tmp;
    DEB(printk("%s ioctl\n", info->dev));

    switch(cmd)
    {
    case READ_INTR_COUNT:
        if(!arg) return -EINVAL;
        if(!access_ok(VERIFY_WRITE, arg, sizeof(int))) /* Plug security hole */
            return -EFAULT;
        spin_lock_irqsave(&device_lock, flags);
        tmp = info->intrs;
        spin_unlock_irqrestore(&device_lock, flags);
        return put_user(tmp, (int *)arg);
    case READ_INTR_CLEAR:
        if(!arg) return -EINVAL;
        if(!access_ok(VERIFY_WRITE, arg, sizeof(int))) /* Plug security hole */
            return -EFAULT;
        spin_lock_irqsave(&device_lock, flags);
        tmp = info->intrs;
        info->intrs = 0;
        spin_unlock_irqrestore(&device_lock, flags);
        return put_user(tmp, (int *)arg);
    case DORK_INTR_COUNT:
        info->intrs = 0;
        break;
    case DISA_INTR_DEVIC:
        info->cntrl &= ~IRQ_ENB;
        outb(info->cntrl, LPT_CTRL_PORT);
        break;
    case ENAB_INTR_DEVIC:
        info->cntrl |= IRQ_ENB;
        outb(info->cntrl, LPT_CTRL_PORT);
        break;
    case GET_DEV_OPENCNT:
        if(!arg) return -EINVAL;
        if(!access_ok(VERIFY_WRITE, arg, sizeof(int))) /* Plug security hole */
            return -EFAULT;
        spin_lock_irqsave(&device_lock, flags);
        tmp = info->open;
        spin_unlock_irqrestore(&device_lock, flags);
        return put_user(tmp, (int *)arg);
        break;
    default:
        return -ESPIPE;
    }
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Open the device.
 */
static int device_open(struct inode *inp, struct file *fp)
{
    DEB(printk("%s open\n", info->dev));
    MOD_INC_USE_COUNT;                         /* Increment usage count */
    info->open++;
    return 0;
} 
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Close the device.
 */
static int device_close(struct inode *inp, struct file *fp)
{
    DEB(printk("%s close\n", info->dev));
    MOD_DEC_USE_COUNT;                         /* Decrement usage count */
    info->open--;
    return 0;
} 
static loff_t device_seek(struct file *a, loff_t b, int c)
{
    return -ESPIPE;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Really complicated interrupt service routine. The kernel does everything
 *  for you as far as the interrupt controller is concerned. You just worry
 *  about your hardware and make certain that it can interrupt again when
 *  your ISR is complete. The printer port will generate new edges without
 *  any attention so we don't have to touch it here.
 */
static void device_isr(int irq, void *p, struct pt_regs *regs)
{
    INFO *inf = (INFO *) p;
    inf->intrs++;
/*
 *  Deliberately scribble on an allocated buffer. It will be present
 *  even in the interrupt context.
 */
    (void)sprintf(inf->buf,"Interrupt IRQ%d (%d) for %s\n", irq, inf->intrs, inf->dev); 
    DEB(printk(inf->buf));
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Initialize and register the module
 */
int __init init_module()
{
    int result;
    spin_lock_init(&device_lock);
    if((info = (INFO *) kmalloc(sizeof(INFO), GFP_KERNEL)) == NULL)
    {
        printk(KERN_CRIT "Can't allocate memory for %s\n", devname);
        return -ENOMEM;
    }
    memset(info, 0x00, sizeof(INFO)); 
    info->dev   = devname;
    info->major = DEV_MAJOR;
    info->irq   = IRQ_USE;
    if((info->buf = (char *) kmalloc(BUF_LEN, GFP_KERNEL)) == NULL)
    {
        printk(KERN_CRIT "Can't allocate memory for %s\n", devname);
        kfree(info);
        return -ENOMEM;
    }
    if((result = register_chrdev(info->major, info->dev, &device_fops)) < 0)
    {
        printk(KERN_ALERT "Can't register %s by major number %d\n", info->dev, info->major);
        kfree(info->buf);
        kfree(info);
        return result;
    }
    if((result = request_irq(info->irq, device_isr, 0, info->dev, info)) < 0)
    {
        printk(KERN_ALERT "Can't allocate IRQ%d for %s\n", info->irq, info->dev);
        (void) unregister_chrdev(info->major, info->dev);
        kfree(info->buf);
        kfree(info);
        return result; 
    }
    info->cntrl = IRQ_ENB;                    /* Turn on device interrupt */
    outb(info->cntrl, LPT_CTRL_PORT);         /* To the control port      */ 
#ifdef DEBUG
    printk(KERN_INFO "Initialization of \"%s\" complete\n", info->dev);
#endif
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Release the module
 */
void cleanup_module()
{
    int result;

    outb(0, LPT_CTRL_PORT);                   /* Turn off interrupt source */ 
    free_irq(info->irq, info);                /* Then free interrupt       */
    if((result = unregister_chrdev(info->major, info->dev)) < 0)
    {
        printk(KERN_ALERT "Can't unregister \"%s\" (%d)\n", info->dev, result);
        return;
    }
    kfree(info->buf);
    kfree(info);
#ifdef DEBUG
    printk(KERN_INFO "%s unregistered. Bye!\n", info->dev);
#endif
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
